/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2021 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.tool.toolbox;

import org.apache.commons.io.IOUtils;
import org.junit.Before;
import org.junit.Test;
import org.openvpms.archetype.test.ArchetypeServiceTest;
import org.openvpms.archetype.test.builder.user.TestUserFactory;
import org.openvpms.component.model.user.User;
import org.openvpms.db.service.DatabaseService;
import org.openvpms.tool.toolbox.config.ConfigProperties;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.test.annotation.DirtiesContext;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.openvpms.tool.toolbox.config.Config.DB_DRIVER;
import static org.openvpms.tool.toolbox.config.Config.DB_PASSWORD;
import static org.openvpms.tool.toolbox.config.Config.DB_URL;
import static org.openvpms.tool.toolbox.config.Config.DB_USERNAME;
import static org.openvpms.tool.toolbox.config.Config.OPENVPMS_KEY;
import static org.openvpms.tool.toolbox.config.Config.REPORTING_DB_PASSWORD;
import static org.openvpms.tool.toolbox.config.Config.REPORTING_DB_URL;
import static org.openvpms.tool.toolbox.config.Config.REPORTING_DB_USERNAME;

/**
 * Tests the {@link Toolbox}.
 * <p/>
 * Note that tool loads its own toolbox-context.xml, which will register its own singletons, replacing those
 * generated by the default applicationContext.xml, hence the @DirtiesContext annotation.
 *
 * @author Tim Anderson
 */
@SuppressWarnings("SpringJavaInjectionPointsAutowiringInspection")
@DirtiesContext(classMode = DirtiesContext.ClassMode.AFTER_EACH_TEST_METHOD)
public class ToolboxTestCase extends ArchetypeServiceTest {

    /**
     * The user factory.
     */
    @Autowired
    private TestUserFactory userFactory;

    /**
     * The password encoder.
     */
    @Autowired
    private PasswordEncoder encoder;

    /**
     * The database service.
     */
    @Autowired
    private DatabaseService databaseService;

    /**
     * Test openvpms home directory.
     */
    private File home;

    /**
     * Sets up the test case.
     *
     * @throws Exception for any error
     */
    @Before
    public void setUp() throws Exception {
        home = initHome("test-home");

        // make sure the database is up-to-date before running tests. This uses dummy migrators
        databaseService.update(() -> {}, () -> {});
    }

    /**
     * Tests the configure --default command.
     */
    @Test
    public void testConfigureDefault() throws IOException {
        File home = initHome("test-home2");
        System.setProperty("openvpms.home", home.getCanonicalPath());

        File conf = new File(home, "conf");
        if (!conf.exists()) {
            assertTrue(conf.mkdirs());
        }

        File file = new File(conf, "openvpms.properties");
        if (file.exists()) {
            assertTrue(file.delete());
        }

        // generate openvpms.properties
        String[] args = {"configure", "--default"};
        assertEquals(0, Toolbox.execute(args));
        assertTrue(file.exists());
        ConfigProperties config = new ConfigProperties(file.toPath());
        assertEquals(config.getDriver(), "com.mysql.cj.jdbc.Driver");
        assertTrue("jdbc:mysql://localhost:3306/openvpms?useSSL=false".equals(config.getUrl())
                   || "jdbc:mysql://localhost:3306/openvpms_dev?useSSL=false".equals(config.getUrl()));
        assertEquals(config.getUsername(), "openvpms");
        assertNotNull(config.getPassword());
        assertEquals(config.getUrl(), config.getReportingUrl());
        assertEquals(config.getUsername(), config.getReportingUsername());
        assertEquals(config.getPassword(), config.getReportingPassword());
        assertNotNull(config.getKey());

        // verify the entries are in the expected order, for better readability
        List<String> keys = new ArrayList<>();
        for (Map.Entry<Object, Object> entry : config.getProperties().entrySet()) {
            keys.add(entry.getKey().toString());
        }
        String[] expected = {DB_DRIVER, DB_URL, DB_USERNAME, DB_PASSWORD,
                             REPORTING_DB_URL, REPORTING_DB_USERNAME,
                             REPORTING_DB_PASSWORD, OPENVPMS_KEY};
        assertArrayEquals(expected, keys.toArray(new String[0]));
    }

    /**
     * Tests the openvpms user --setpassword command.
     */
    @Test
    public void testSetPassword() {
        User user = userFactory.createUser();
        String[] args = {"user", "--setpassword", user.getUsername(), "-p", "foo"};
        assertEquals(0, Toolbox.execute(args));
        user = get(user);
        String password = getBean(user).getString("password");
        assertNotEquals("foo", password);
        assertTrue(encoder.matches("foo", password));
    }

    /**
     * Tests the openvpms user --enable command.
     */
    @Test
    public void testEnable() {
        User user = userFactory.newUser().active(false).build();
        assertFalse(user.isActive());
        String[] args1 = {"user", "--enable", user.getUsername()};
        assertEquals(0, Toolbox.execute(args1));
        user = get(user);
        assertTrue(user.isActive());

        String[] args2 = {"user", "--enable", "xsomenonexistentuser"};
        assertEquals(1, Toolbox.execute(args2));
    }

    /**
     * Tests the openvpms user --disable command.
     */
    @Test
    public void testDisable() {
        User user = userFactory.createUser();
        assertTrue(user.isActive());
        String[] args1 = {"user", "--disable", user.getUsername()};
        assertEquals(0, Toolbox.execute(args1));
        user = get(user);
        assertFalse(user.isActive());

        String[] args2 = {"user", "--disable", "xsomenonexistentuser"};
        assertEquals(1, Toolbox.execute(args2));
    }

    /**
     * Verifies that the openvpms.war is produced by the war argument.
     *
     * @throws Exception for any error
     */
    @Test
    public void testWar() throws Exception {
        String[] args = {"war"};
        Path war = Paths.get(home.getPath(), "webapps", "openvpms.war");
        if (Files.exists(war)) {
            Files.delete(war);
        }
        assertFalse(Files.exists(war));
        assertEquals(0, Toolbox.execute(args));
        assertTrue(Files.exists(war));
    }

    /**
     * Verifies that the the war name can be specified, and that logj42.xml and web.xml are updated.
     *
     * @throws Exception for any error
     */
    @Test
    public void testWarName() throws Exception {
        String[] args = {"war", "--name", "demo"};
        Path war = Paths.get(home.getPath(), "webapps", "demo.war");
        if (Files.exists(war)) {
            Files.delete(war);
        }
        assertFalse(Files.exists(war));
        assertEquals(0, Toolbox.execute(args));
        assertTrue(Files.exists(war));

        boolean foundLog4j = false;
        boolean foundWebXML = false;
        try (JarFile jar = new JarFile(war.toFile())) {
            for (Enumeration<JarEntry> entries = jar.entries(); entries.hasMoreElements(); ) {
                JarEntry entry = entries.nextElement();
                String name = entry.getName();
                if (name.equals("WEB-INF/classes/log4j2.xml")) {
                    foundLog4j = true;
                    checkLog4J(jar.getInputStream(entry));
                    if (foundWebXML) {
                        break;
                    }
                } else if (name.equals("WEB-INF/web.xml")) {
                    foundWebXML = true;
                    checkWebXML(jar.getInputStream(entry));
                    if (foundLog4j) {
                        break;
                    }
                }
            }
        }
        assertTrue(foundLog4j);
        assertTrue(foundWebXML);
    }

    /**
     * Returns a home directory relative to the target dir, and sets the openvpms.home system property.
     *
     * @param dir the directory name
     * @return the home directory
     */
    private File initHome(String dir) throws IOException {
        String path = Toolbox.class.getProtectionDomain().getCodeSource().getLocation().getFile();
        File home = new File(path, "../" + dir);
        System.setProperty("openvpms.home", home.getCanonicalPath());
        return home;
    }

    /**
     * Verifies that the log4j files are named demo.log and demo-full.log.
     *
     * @param stream the jar entry stream
     * @throws IOException for any I/O exception
     */
    private void checkLog4J(InputStream stream) throws IOException {
        int matches = 0;
        List<String> lines = IOUtils.readLines(stream, StandardCharsets.UTF_8);
        for (String line : lines) {
            if (line.contains("RollingFile name=\"fullout\" fileName=\"${sys:catalina.base}/logs/demo-full.log\"")) {
                matches++;
            } else if (line.contains("filePattern=\"${sys:catalina.base}/logs/demo-full.log.%i\"")) {
                matches++;
            } else if (line.contains("<RollingFile name=\"fileout\" fileName=\"${sys:catalina.base}/logs/demo.log\"")) {
                matches++;
            } else if (line.contains("filePattern=\"${sys:catalina.base}/logs/demo.log.%i\"")) {
                matches++;
            }
        }
        assertEquals(4, matches);
    }

    /**
     * Verifies that the web.xml contains an {@code <param-value>demo</param-value>}.
     * <p/>
     * This is used for the webAppRootKey.
     *
     * @param stream the jar entry stream
     * @throws IOException for any I/O exception
     */
    private void checkWebXML(InputStream stream) throws IOException {
        int matches = 0;
        List<String> lines = IOUtils.readLines(stream, StandardCharsets.UTF_8);
        for (String line : lines) {
            if (line.contains("<param-value>demo</param-value>")) {
                matches++;
            }
        }
        assertEquals(1, matches);
    }

}
