/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2021 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.workspace.admin.eftpos;

import echopointng.HttpPaneEx;
import org.openvpms.eftpos.internal.event.EFTPOSEventMonitor;
import org.openvpms.eftpos.service.WebTerminalRegistrar;
import org.openvpms.eftpos.terminal.Terminal;
import org.openvpms.eftpos.terminal.TerminalStatus;
import org.openvpms.web.echo.dialog.InformationDialog;
import org.openvpms.web.echo.dialog.ModalDialog;
import org.openvpms.web.echo.dialog.PopupDialogListener;
import org.openvpms.web.echo.util.ApplicationInstanceConsumer;
import org.openvpms.web.system.ServiceHelper;

/**
 * Terminal registration dialog.
 *
 * @author Tim Anderson
 */
public class TerminalRegistrationDialog extends ModalDialog {

    /**
     * The terminal to register.
     */
    private final Terminal terminal;

    /**
     * The monitor for EFTPOS events.
     */
    private final EFTPOSEventMonitor monitor;

    /**
     * The listener.
     */
    private ApplicationInstanceConsumer<TerminalStatus> listener;


    /**
     * Constructs a {@link TerminalRegistrationDialog}.
     *
     * @param terminal  the terminal to register
     * @param registrar the terminal registrar
     */
    public TerminalRegistrationDialog(Terminal terminal, WebTerminalRegistrar registrar) {
        super("Register EFTPOS Terminal", CANCEL, null);
        this.terminal = terminal;
        setContentWidth(registrar.getWidth());
        setContentHeight(registrar.getHeight());
        monitor = ServiceHelper.getBean(EFTPOSEventMonitor.class);
        HttpPaneEx frame = new HttpPaneEx(registrar.getUrl());
        getLayout().add(frame);
    }

    /**
     * Show the window.
     */
    @Override
    public void show() {
        super.show();
        listener = new ApplicationInstanceConsumer<>(this::onCompleted);
        monitor.addTerminalListener(terminal, listener);
    }

    /**
     * Disposes of this dialog.
     */
    @Override
    public void dispose() {
        super.dispose();
        if (listener != null) {
            listener.dispose();
        }
    }

    private void onCompleted(TerminalStatus status) {
        monitor.removeTerminalListener(terminal, listener);
        // only display the notification if the dialog is still popped up
        if (getParent() != null) {
            if (status.isRegistered()) {
                InformationDialog.show("EFTPOS Terminal", "EFTPOS terminal " + terminal.getName()
                                                          + " has been registered",
                                       new PopupDialogListener() {
                                           @Override
                                           public void onOK() {
                                               TerminalRegistrationDialog.this.close(OK_ID);
                                           }
                                       });

            } else {
                InformationDialog.show("EFTPOS Terminal",
                                       "Failed to register EFTPOS terminal " + terminal.getName() + ": "
                                       + status.getMessage(),
                                       new PopupDialogListener() {
                                           @Override
                                           public void onOK() {
                                               TerminalRegistrationDialog.this.close(CANCEL_ID);
                                           }
                                       });
            }
        }
    }
}
